#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function performs a contrast normalization on an image.  The mean   */
/*   graylevel is shifted, and the graylevels are stretched to result in      */
/*   the specified standard deviation from the mean.                          */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void CONTNORM (
IMAGE  *in,     /*  I   Pointer to the input image.                           */
PIXEL  mean,    /*  I   Mean graylevel in the output image.                   */
double sigma,   /*  I   Standard deviation in the output image.               */
uint32_t inc,   /*  I   Line and pixel subsample increment.                   */
IMAGE  **out    /*  O   Address of a pointer to the output image.             */
                /*      The input image may be overwritten (out = &in).       */
/*----------------------------------------------------------------------------*/
) { register uint32_t i, j, k;
    char   msg[SLEN];
    double oldsigma, gain, bias;
    PIXEL  sum, sumsq, oldmean;

    if (TIMES) TIMING(T_CONTNORM);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"CONTNORM");
    }

    /* check input */
    if (!CHECKIMG(in)) {
        MESSAGE('E'," Can't identify image.");
        goto the_end;
    } else if (inc <= 0) {
        MESSAGE('E'," Subsample increment must be greater than zero.");
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");  
        sprintf(msg," Input image:                          %s",in->text);
        MESSAGE('I',msg);
        sprintf(msg," Graylevel mean:                     %12.4e",mean);
        MESSAGE('I',msg);
        sprintf(msg," Graylevel standard deviation:       %12.4e",sigma);
        MESSAGE('I',msg);
        sprintf(msg," Line and pixel subsample increment:   %lu",inc);
        MESSAGE('I',msg);
        MESSAGE('I'," ...............");
    }

    /* create image of appropriate size */
    if (!CHECKIMG(*out)) GETMEM(in->nbnd,in->nlin,in->npix,out);
    if (!*out) goto the_end;

    /* compute statistics */
    for (sum=sumsq=0.0,i=0; i<in->nbnd; i++) {
        for (j=0; j<in->nlin; j+=inc) {
            for (k=0; k<in->npix; k+=inc) {
                sum   += in->data[i][j][k];
                sumsq += in->data[i][j][k]*in->data[i][j][k];
            }
        }
    }    
    oldmean  = (PIXEL)((double)sum/((double)in->nbnd*(double)(in->nlin/inc+(in->nlin%inc ? 1:0))*(double)(in->npix/inc+(in->npix%inc ? 1:0))));
    oldsigma = sqrt((double)sumsq/((double)in->nbnd*(double)(in->nlin/inc+(in->nlin%inc ? 1:0))*(double)(in->npix/inc+(in->npix%inc ? 1:0)))-(double)oldmean*(double)oldmean);

    /* calculate parameters for stretch */
    gain = (oldsigma == 0.0 ? 1.0 : sigma / oldsigma);
    bias = (double)mean - gain*(double)oldmean;
    
    /* perform stretch */
    for (i=0; i<in->nbnd; i++) {
        for (j=0; j<in->nlin; j++) {
            for (k=0; k<in->npix; k++) {
                (*out)->data[i][j][k] = (PIXEL)(gain*(double)in->data[i][j][k] + bias);
            }
        }
    }

    the_end:
    if (TIMES) TIMING(T_EXIT);
}
