#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function computes the band-to-band covariance and correlation       */
/*   matrices of an image.                                                    */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void COVAR (
IMAGE  *in,     /*  I   Pointer to the image.                                 */
uint32_t inc,   /*  I   Pixels/line and line increment.                       */
double *cov,    /*  O   Pointer to the covariance matrix.                     */
double *cor     /*  O   Pointer to the correlation matrix.                    */
/*----------------------------------------------------------------------------*/
) { register uint64_t i, j, k, l;
    uint64_t total;
    char   msg[SLEN];
    double *sum=NULL, *sumsq=NULL, count, pinc, psum;

    if (TIMES) TIMING(T_COVAR);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"COVAR");
    }

    /* check input */
    if (!CHECKIMG(in)) {
        MESSAGE('E'," Can't identify image.");
        goto the_end;
    } else if (inc <= 0) {
        MESSAGE('E'," Subsample increment must be greater than zero.");
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Input image:                          %s",in->text);
        MESSAGE('I',msg);
        sprintf(msg," Line and pixel subsample increment:   %lu",inc);
        MESSAGE('I',msg);
        MESSAGE('I'," ...............");
    }

    /* allocate space for summation arrays */
    total = (uint64_t)in->nbnd * (uint64_t)sizeof(double);
    if (total > (uint64_t)SIZE_MAX  ||  (sum=(double *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }
    total = (uint64_t)in->nbnd * (uint64_t)in->nbnd * (uint64_t)sizeof(double);
    if (total > (uint64_t)SIZE_MAX  ||  (sumsq=(double *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }

    /* initialize progress indicator */
    if (LINES  &&  !PROGRESS(psum=0.0)) goto the_end;
    pinc = 2.0/(double)(in->nbnd+1)/(double)in->nbnd/(double)(in->nlin/inc + (in->nlin%inc?1:0));

    /* initialize */
    for (i=0; i<in->nbnd; i++) {
        sum[i] = 0.0;
        for (j=0; j<in->nbnd; j++) {
            sumsq[i*in->nbnd+j] = 0.0;
        }
    }

    /* compute the sums */
    for (i=0; i<in->nbnd; i++) {
        for (j=0; j<in->nlin; j+=inc) {
            for (k=0; k<in->npix; k+=inc) {
                sum[i]              += (double)in->data[i][j][k];
                sumsq[i*in->nbnd+i] += (double)in->data[i][j][k] * (double)in->data[i][j][k];
            }
            if (LINES  &&  !PROGRESS(psum+=pinc)) goto the_end;
        }
    }
    for (i=0; i<in->nbnd-1; i++) {
        for (l=i+1; l<in->nbnd; l++) {
            for (j=0; j<in->nlin; j+=inc) {
                for (k=0; k<in->npix; k+=inc) {
                    sumsq[i*in->nbnd+l] = sumsq[l*in->nbnd+i] += (double)in->data[i][j][k] * (double)in->data[l][j][k];
                }
                if (LINES  &&  !PROGRESS(psum+=pinc)) goto the_end;
            }
        }
    }

    /* calculate the covariance matrix */
    count = (double)(in->nlin/inc+(in->nlin%inc ? 1:0)) * (double)(in->npix/inc+(in->npix%inc ? 1:0));
    for (i=0; i<in->nbnd; i++) {
        for (j=0; j<in->nbnd; j++) {
            cov[i*in->nbnd+j] = (sumsq[i*in->nbnd+j]-sum[i]*sum[j]/count)/(count-1.0);
        }
    }

    /* calculate the correlation matrix */
    for (i=0; i<in->nbnd; i++) {
        for (j=0; j<in->nbnd; j++) {
            cor[i*in->nbnd+j] = cov[i*in->nbnd+j]/sqrt(cov[i*in->nbnd+i]*cov[j*in->nbnd+j]);
        }
    }

    /* output both matrices */
    MESSAGE('I',"");
    MESSAGE('I'," Covariance Matrix");
    for (i=0; i<in->nbnd; i++) {
        sprintf(msg," ");
        for (j=0; j<min(3,in->nbnd); j++) {
            sprintf(msg+strlen(msg),"%12.4e",cov[i*in->nbnd+j]);
        }
        MESSAGE('I',msg);
    }
    MESSAGE('I',"");
    MESSAGE('I'," Correlation Matrix");
    for (i=0; i<in->nbnd; i++) {
        sprintf(msg," ");
        for (j=0; j<min(3,in->nbnd); j++) {
            sprintf(msg+strlen(msg),"%12.4e",cor[i*in->nbnd+j]);
        }
        MESSAGE('I',msg);
    }

    the_end:
    if (sum)   free(sum);
    if (sumsq) free(sumsq);
    if (TIMES) TIMING(T_EXIT);
}
