#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include "sadie.h"

#if defined SADIE_MACOS
#include <Files.h>
#endif

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function reads a SADIE image from disk into main memory.            */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void DISK2IMG (
unsigned char *name, /*  I   String, containing the name of the image file.   */
short ref,           /*  I   Reference number of the image file.              */
uint64_t size,       /*  I   Size, in bytes, of the image file.               */
IMAGE **out          /*  O   Address of a pointer to the image.               */
/*----------------------------------------------------------------------------*/
) { static char array[2] = { 0x00, 0x01 };
    register uint32_t i, j, k;
    uint64_t total;
    char msg[SLEN];
    long nbytes;
    IMAGE img;
    IMAGEv0 imgv0;
#if defined SADIE_UNIX
    FILE *fp = NULL;
#elif defined SADIE_MACOS
    short error;
#endif

    if (TIMES) TIMING(T_DISK2IMG);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"DISK2IMG");
    }

#if defined SADIE_UNIX
    /* open image file */
    if (!(fp=fopen(name,"r"))) {
        sprintf(msg," Can't open file %s.",name);
        MESSAGE('E',msg);
        goto the_end;
    }
#endif

    /* read image header from file */
    nbytes = (long)(total = (uint64_t)sizeof(img) - (uint64_t)sizeof(img.data));
#if defined SADIE_UNIX
    if (fread(&img,(size_t)total,1,fp) != 1) {
        sprintf(msg," Can't read image header from file %s.",name);
        MESSAGE('E',msg);
        goto the_end;
    }
#elif defined SADIE_MACOS
    if ((error=FSRead(ref,&nbytes,&img)) != noErr) {
        sprintf(msg," Can't read image header from file %s.\rFile Manager Error Code: %d.",name,error);
        MESSAGE('E',msg);
        goto the_end;
    }
#endif

    /* swap bytes, if necessary */
    if (*(short *)array != 1) {
        SWAP2((BYTE *)&img.fileformat);
        SWAP2((BYTE *)&img.fileversion);
        SWAP2((BYTE *)&img.application);
        SWAP2((BYTE *)&img.appversion);
        SWAP2((BYTE *)&img.dataformat);
        SWAP2((BYTE *)&img.datapacking);
        SWAP4((BYTE *)&img.npix);
        SWAP4((BYTE *)&img.nlin);
        SWAP4((BYTE *)&img.nbnd);
        SWAP4((BYTE *)&img.ntvl);
        SWAP4((BYTE *)&img.nbit);
        SWAP4((BYTE *)&img.gmin);
        SWAP4((BYTE *)&img.gmax);
    }

    /* check for magic number, assume version 0 image if not present */
    if (img.fileformat != FILE_FORMAT) {
        nbytes = (long)(total = (uint64_t)sizeof(imgv0) - (uint64_t)sizeof(imgv0.data));
#if defined SADIE_UNIX
        if (fseek(fp,0,SEEK_SET) != 0  ||  fread(&imgv0,(size_t)total,1,fp) != 1) {
            sprintf(msg," Can't read image header from file %s.",name);
            MESSAGE('E',msg);
            goto the_end;
        }
#elif defined SADIE_MACOS
        if ((error=SetFPos(ref,fsFromStart,0)) != noErr  ||  (error=FSRead(ref,&nbytes,&imgv0)) != noErr) {
            sprintf(msg," Can't read image header from file %s.\rFile Manager Error Code: %d.",name,error);
            MESSAGE('E',msg);
            goto the_end;
        }
#endif
        if (*(short *)array != 1) {
            SWAP2((BYTE *)&imgv0.nbnd);
            SWAP2((BYTE *)&imgv0.nlin);
            SWAP2((BYTE *)&imgv0.npix);
            SWAP2((BYTE *)&imgv0.nbit);
        }
        if (imgv0.nbnd > 0  &&  imgv0.nlin > 0  &&  imgv0.npix > 0  &&  imgv0.nbit > 0) {
            img.fileformat  = FILE_FORMAT;
            img.fileversion = 0;
            img.npix        = imgv0.npix;
            img.nlin        = imgv0.nlin;
            img.nbnd        = imgv0.nbnd;
            img.nbit        = imgv0.nbit;
            img.gmin        = imgv0.gmin;
            img.gmax        = imgv0.gmax;
        }
    }

    /* check that the number of bytes in the file matches the image size */
    nbytes = (long)(total = (uint64_t)ceil(((double)img.nbnd*(double)img.nlin*(double)img.npix*(double)img.nbit)/(double)SBPB));
    if (size > 0  &&  size != total+(uint64_t)(img.fileversion == 0 ? sizeof(imgv0)-sizeof(imgv0.data) : sizeof(img)-sizeof(img.data))) {
        sprintf(msg," Size of file %s does not match image size:",name);
        MESSAGE('E',msg);
        sprintf(msg," File size = %llu, image size = %llu.",size,total+(uint64_t)(img.fileversion == 0 ? sizeof(imgv0)-sizeof(imgv0.data) : sizeof(img)-sizeof(img.data)));
        MESSAGE('E',msg);
        sprintf(msg," Nbnd = %lu, nlin = %lu, npix = %lu, nbit = %lu.",img.nbnd,img.nlin,img.npix,img.nbit);
        MESSAGE('E',msg);
        goto the_end;
    }

    /* check number of bits per pixel */
    if (img.nbit != SBIT   &&   (SBIT < img.nbit  ||  img.nbit > sizeof(long)*SBPB)) {
        sprintf(msg," Can't unpack from %lu bits/pixel.",img.nbit);
        MESSAGE('E',msg);
        goto the_end;
    }

    /* create image of appropriate size */
    if (!CHECKIMG(*out)) GETMEM(img.nbnd,img.nlin,img.npix,out);
    if (!*out) goto the_end;

    /* read the image data from file */
#if defined SADIE_UNIX
    if (fread((*out)->data[0][0],(size_t)total,1,fp) != 1) {
        sprintf(msg," Can't read image data from file %s.",name);
        MESSAGE('E',msg);
    }
#elif defined SADIE_MACOS
    if ((error=FSRead(ref,&nbytes,(*out)->data[0][0])) != noErr) {
        sprintf(msg," Can't read image data from file %s.\rFile Manager Error Code: %d.",name,error);
        MESSAGE('E',msg);
        goto the_end;
    }
#endif
    if (img.nbit < SBIT) {
        UNPACK(img.nbit,(uint64_t)img.nbnd*(uint64_t)img.nlin*(uint64_t)img.npix,(*out)->data[0][0]);
    } else if (img.fileversion > 0  &&  img.nbit == SBIT  &&  *(short *)array != 1) {
        for (i=0; i<img.nbnd; i++) {
            for (j=0; j<img.nlin; j++) {
                for (k=0; k<img.npix; k++) {
                    SWAP4((BYTE *)&(*out)->data[i][j][k]);
                }
            }
        }
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Input file:                                 %s",name);
        MESSAGE('I',msg); 
        sprintf(msg," Input format:                               0x%X, version %d",img.fileformat,img.fileversion);
        MESSAGE('I',msg); 
        sprintf(msg," Header size:                                %lu",(unsigned long)(img.fileversion == 0 ? sizeof(imgv0)-sizeof(imgv0.data) : sizeof(img)-sizeof(img.data)));
        MESSAGE('I',msg);
        sprintf(msg," Number of bands, lines, pixels:             %lu, %lu, %lu",img.nbnd,img.nlin,img.npix);
        MESSAGE('I',msg);
        sprintf(msg," Number of bits/pixel/band (image file):     %lu",img.nbit);
        MESSAGE('I',msg);
        sprintf(msg," Graylevel minimum, maximum:                 %10.4e, %10.4e",img.gmin,img.gmax);
        MESSAGE('I',msg);
        MESSAGE('I'," ...............");
    }

    /* close image file */
    the_end:
#if defined SADIE_UNIX
    if (fp) {
        if (fclose(fp)) {
            sprintf(msg," Can't close file %s.",name);
            MESSAGE('W',msg);
        }
    }
#endif
    if (TIMES) TIMING(T_EXIT);
}
