#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1989 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This low-level function computes a discrete one-dimensional Hartley      */
/*   transform.  The Fourier transform may also be easily computed from       */
/*   the results.                                                             */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void FHT1D (
short  option,  /*  I   Switch, indicating the direction of the transform:    */
                /*      FORWARD   -   forward Hartley transform is computed.  */
                /*      INVERSE   -   inverse Hartley transform is computed.  */
uint32_t size,  /*  I   Number of data points.                                */
uint32_t p,     /*  I   Logarithm (base 2) of the number of data points.      */
uint32_t *a,    /*  I   Permutation index array[p/2+p%2].                     */
uint32_t *pwr,  /*  I   Array containing powers of two.                       */
double *sine,   /*  I   Array containing sine function values.                */
double *tang,   /*  I   Array containing tangent function values.             */
double *array   /* I/O  Pointer to the data array[size].                      */
/*----------------------------------------------------------------------------*/
) { register uint32_t o, n, m, l, k, j, i;
    register double t, u, v, w, x, y;

    if (TIMES) TIMING (T_FHT1D);

    a[0] = 0;
    a[1] = 1;
    for (i=2; i<=p/2+p%2; i++) {
        for (j=0; j<pwr[i-1]; j++) {
            a[j+pwr[i-1]] = (a[j] *= 2) + 1;
        }
    }
    for (i=1; i<pwr[p/2]; i++) {
        k = i;
        l = m = pwr[p/2]*a[i];
        t        = array[k];
        array[k] = array[m];
        array[m] = t;
        for (j=1; j<=a[i]-1; j++) {
            k += pwr[p/2];
            m  = l + a[j];
            t        = array[k];
            array[k] = array[m];
            array[m] = t;
        }
    }
    for (i=0; i<size-1; i+=2) {
        t = array[i] + array[i+1];
        u = array[i] - array[i+1];
        array[i  ] = t;
        array[i+1] = u;
    }
    for (i=0; i<size-3; i+=4) {
        t = array[i  ] + array[i+2];
        u = array[i+1] + array[i+3];
        v = array[i  ] - array[i+2];
        w = array[i+1] - array[i+3];
        array[i  ] = t;
        array[i+1] = u;
        array[i+2] = v;
        array[i+3] = w;
    }
    for (i=2; i<p; i++) {
        for (j=0; j<size; j+=pwr[i+1]) {
            t = array[j] + array[j+pwr[i]];
            u = array[j] - array[j+pwr[i]];
            array[j       ] = t;
            array[j+pwr[i]] = u;
            l = j + 1;
            m = j + pwr[i] - 1;
            n = l + pwr[i];
            o = m + pwr[i];
            for (k=pwr[p-i-1]; k<=size/4; k+=pwr[p-i-1]) {
                t = array[n] + tang[k]*array[o];
                x = array[o] - sine[k]*t;
                y = t        + tang[k]*x;
                t = array[l] + y;
                u = array[m] - x;
                v = array[l] - y;
                w = array[m] + x;
                array[l++] = t;
                array[m--] = u;
                array[n++] = v;
                array[o--] = w;
            }
        }
    }

    if (option == FORWARD) {
        for (i=0; i<size; i++) {
            array[i] /= (double)size;
        }
    }

    the_end:
    if (TIMES) TIMING(T_EXIT);
}

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1989 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function computes the Hartley transform of an image.                */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Background Information-----------------------------------------------------*/
/*                                                                            */
/*   Bracewell, R.N.:                                                         */
/*   "The Hartley Transform."                                                 */
/*   Oxford University Press, New York, 1986                                  */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void FHT2D (
IMAGE *in,      /*  I   Pointer to the input image.                           */
short option,   /*  I   Switch, indicating the direction of the transform:    */
                /*      FORWARD   -   forward Hartley transform is computed.  */
                /*      INVERSE   -   inverse Hartley transform is computed.  */
IMAGE **out     /*  O   Address of a pointer to the output image.             */
/*----------------------------------------------------------------------------*/
) { register int64_t i, j, k;
    uint32_t p, *a=NULL, *pwr=NULL;
    uint64_t total;
    char   msg[SLEN];
    double *sine=NULL, *tang=NULL, *array=NULL, pinc, psum;
    PIXEL  t, u, v, w;
    IMAGE  *img;

    if (TIMES) TIMING(T_FHT2D); 
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"FHT2D");
    }

    /* check input */
    if (!CHECKIMG(in)) {
        MESSAGE('E'," Can't identify image.");
        goto the_end;
    } else if (in->nlin != 1<<rnd(log10((double)in->nlin)/log10(2.0))) {
        MESSAGE('E'," Number of lines must be a power of two.");
        goto the_end;
    } else if (in->npix != 1<<rnd(log10((double)in->npix)/log10(2.0))) {
        MESSAGE('E'," Number of pixels/line must be a power of two.");
        goto the_end;
    } else if (option == FORWARD  &&  in->npix > UINT32_MAX/2) {
        MESSAGE('E'," Image size (number of pixels/line) too large.");
        goto the_end;
    } else if (option != FORWARD  &&  option != INVERSE) {
        MESSAGE('E'," Transformation option must be either FORWARD or INVERSE.");
        goto the_end;
    } else if (option == FORWARD  &&  in->nlin != in->npix   ||   option == INVERSE  &&  in->nlin != in->npix/2) {
        MESSAGE('E'," Number of lines must be equal to number of pixels/line.");
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Input image:   %s",in->text);
        MESSAGE('I',msg);
        if (option == FORWARD) {
            MESSAGE('I'," Forward FHT");
        } else if (option == INVERSE) {
            MESSAGE('I'," Inverse FHT");
        }
        MESSAGE('I'," ...............");
    }

    /* create image of appropriate size */
    if (!CHECKIMG(*out)) GETMEM(in->nbnd,in->nlin,option==FORWARD ? 2*in->npix:in->npix/2,out);
    if (!*out) goto the_end;
    img = option == FORWARD  ?  in : *out;

    /* special case */
    if (img->nlin == 1) {
        (*out)->data[0][0][0] = in->data[0][0][0];
        if (option == FORWARD) {
            (*out)->data[0][0][1] = 0.0;
        }
        goto the_end;
    }

    /* allocate permutation index array */
    i = max(img->nlin,img->npix);
    p = rnd(log10((double)i)/log10(2.0));
    total = (uint64_t)(1<<p/2+p%2) * (uint64_t)sizeof(uint32_t);
    if (total > (uint64_t)SIZE_MAX  ||  (a=(uint32_t *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }

    /* allocate and initialize table with powers of two */
    total = (uint64_t)(p+1) * (uint64_t)sizeof(uint32_t);
    if (total > (uint64_t)SIZE_MAX  ||  (pwr=(uint32_t *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }
    for (j=0; j<=p; j++) {
        pwr[j] = 1<<j;
    }

    /* allocate and initialize trigonometric tables */
    total = (uint64_t)(i/4+1) * (uint64_t)sizeof(double);
    if (total > (uint64_t)SIZE_MAX  ||  (sine=(double *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }
    total = (uint64_t)(i/4+1) * (uint64_t)sizeof(double);
    if (total > (uint64_t)SIZE_MAX  ||  (tang=(double *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }
    for (j=0; j<=i/4; j++) {
        sine[j] = sin((2.0*PI*(double)j)/(double)i);
        tang[j] = tan((    PI*(double)j)/(double)i);
    }
            
    /* allocate array to hold one row/column of pixels */
    total = (uint64_t)i * (uint64_t)sizeof(double);
    if (total > (uint64_t)SIZE_MAX  ||  (array=(double *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }

    /* initialize progress indicator */
    if (LINES  &&  !PROGRESS(psum=0.0)) goto the_end;
    pinc = 1.0/(double)img->nbnd/(double)(img->nlin+img->npix);

    /* Hartley transform bands */
    for (i=0; i<(*out)->nbnd; i++) {

        if (option == FORWARD) {
            /* fold */
            for (j=0; j<(*out)->nlin/2; j++) {
                for (k=0; k<(*out)->npix/4; k++) {
                    (*out)->data[i][j][k] = in->data[i][in->nlin/2+j][in->npix/2+k];
                    (*out)->data[i][(*out)->nlin/2+j][k] = in->data[i][j][in->npix/2+k];
                    (*out)->data[i][j][(*out)->npix/4+k] = in->data[i][in->nlin/2+j][k];
                    (*out)->data[i][(*out)->nlin/2+j][(*out)->npix/4+k] = in->data[i][j][k];
                }
            }
        } else {
            /* compute Hartley from Fourier coefficients and fold */
            for (j=0; j<(*out)->nlin/2; j++) {
                for (k=0; k<(*out)->npix/2; k++) {
                    (*out)->data[i][j][k] = in->data[i][in->nlin/2+j][in->npix/2+2*k] - in->data[i][in->nlin/2+j][in->npix/2+2*k+1];
                    (*out)->data[i][(*out)->nlin/2+j][k] = in->data[i][j][in->npix/2+2*k] - in->data[i][j][in->npix/2+2*k+1];
                    (*out)->data[i][j][(*out)->npix/2+k] = in->data[i][in->nlin/2+j][2*k] - in->data[i][in->nlin/2+j][2*k+1];
                    (*out)->data[i][(*out)->nlin/2+j][(*out)->npix/2+k] = in->data[i][j][2*k] - in->data[i][j][2*k+1];
                }
            }
        }

        /* Hartley transform rows */
        for (j=0; j<img->nlin; j++) {
            for (k=0; k<img->npix; k++) {
                array[k] = (double)(*out)->data[i][j][k];
            }
            FHT1D(option,img->npix,p,a,pwr,sine,tang,array);
            for (k=0; k<img->npix; k++) {
                (*out)->data[i][j][k] = (PIXEL)array[k];
            }
            if (LINES  &&  !PROGRESS(psum+=pinc)) goto the_end;
        }

        /* Hartley transform columns */
        for (k=img->npix-1; k>=0; k--) {
            for (j=0; j<img->nlin; j++) {
                array[j] = (double)(*out)->data[i][j][k];
            }
            FHT1D(option,img->nlin,p,a,pwr,sine,tang,array);
            if (option == FORWARD) {
                for (j=0; j<img->nlin; j++) { 
                    (*out)->data[i][j][2*k] = (PIXEL)array[j];
                }
            } else {
                for (j=0; j<img->nlin; j++) { 
                    (*out)->data[i][j][k] = (PIXEL)array[j];
                }
            }
            if (LINES  &&  !PROGRESS(psum+=pinc)) goto the_end;
        }

        if (option == FORWARD) {
            /* compute Fourier from Hartley coefficients */
            (*out)->data[i][0][1] = (*out)->data[i][(*out)->nlin/2][1] = (*out)->data[i][0][(*out)->npix/2+1] = (*out)->data[i][(*out)->nlin/2][(*out)->npix/2+1] = 0.0;
            for (k=2; k<(*out)->npix/2; k+=2) {
                (*out)->data[i][0][(*out)->npix-k+1] = -((*out)->data[i][0][k+1] = ((*out)->data[i][0][(*out)->npix-k]-(*out)->data[i][0][k]) / 2.0);
                (*out)->data[i][0][(*out)->npix-k  ] =  ((*out)->data[i][0][k  ] = ((*out)->data[i][0][(*out)->npix-k]+(*out)->data[i][0][k]) / 2.0);
                (*out)->data[i][(*out)->nlin/2][(*out)->npix-k+1] = -((*out)->data[i][(*out)->nlin/2][k+1] = ((*out)->data[i][(*out)->nlin/2][(*out)->npix-k]-(*out)->data[i][(*out)->nlin/2][k]) / 2.0);
                (*out)->data[i][(*out)->nlin/2][(*out)->npix-k  ] =  ((*out)->data[i][(*out)->nlin/2][k  ] = ((*out)->data[i][(*out)->nlin/2][(*out)->npix-k]+(*out)->data[i][(*out)->nlin/2][k]) / 2.0);
            }
            for (j=1; j<(*out)->nlin/2; j++) {
                (*out)->data[i][(*out)->nlin-j][1] = -((*out)->data[i][j][1] = ((*out)->data[i][(*out)->nlin-j][0]-(*out)->data[i][j][0]) / 2.0);
                (*out)->data[i][(*out)->nlin-j][0] =  ((*out)->data[i][j][0] = ((*out)->data[i][(*out)->nlin-j][0]+(*out)->data[i][j][0]) / 2.0);
                (*out)->data[i][(*out)->nlin-j][(*out)->npix/2+1] = -((*out)->data[i][j][(*out)->npix/2+1] = ((*out)->data[i][(*out)->nlin-j][(*out)->npix/2]-(*out)->data[i][j][(*out)->npix/2]) / 2.0);
                (*out)->data[i][(*out)->nlin-j][(*out)->npix/2  ] =  ((*out)->data[i][j][(*out)->npix/2  ] = ((*out)->data[i][(*out)->nlin-j][(*out)->npix/2]+(*out)->data[i][j][(*out)->npix/2]) / 2.0);
                for (k=2; k<(*out)->npix/2; k+=2) {
                    t = (*out)->data[i][(*out)->nlin-j][(*out)->npix-k];
                    u = (*out)->data[i][j][k];
                    v = (*out)->data[i][j][(*out)->npix-k];
                    w = (*out)->data[i][(*out)->nlin-j][k];
                    (*out)->data[i][(*out)->nlin-j][(*out)->npix-k+1] = -((*out)->data[i][j][k+1] = (t-u) / 2.0);
                    (*out)->data[i][j][(*out)->npix-k  ] =  ((*out)->data[i][(*out)->nlin-j][k  ] = (t+u) / 2.0);
                    (*out)->data[i][j][(*out)->npix-k+1] = -((*out)->data[i][(*out)->nlin-j][k+1] = (v-w) / 2.0);
                    (*out)->data[i][(*out)->nlin-j][(*out)->npix-k  ] =  ((*out)->data[i][j][k  ] = (v+w) / 2.0);
                }
            }
        } else {
            /* adjust Hartley coefficients */
            for (j=1; j<(*out)->nlin/2; j++) {
                for (k=1; k<(*out)->npix/2; k++) {
                    t = ((*out)->data[i][j][k] - (*out)->data[i][(*out)->nlin-j][k] - (*out)->data[i][j][(*out)->npix-k] + (*out)->data[i][(*out)->nlin-j][(*out)->npix-k]) / 2.0;
                    (*out)->data[i][j][k]                           -= t;
                    (*out)->data[i][(*out)->nlin-j][k]              += t;
                    (*out)->data[i][j][(*out)->npix-k]              += t;
                    (*out)->data[i][(*out)->nlin-j][(*out)->npix-k] -= t;
                }
            }
        }
 
        /* unfold */
        for (j=0; j<(*out)->nlin/2; j++) {
            for (k=0; k<(*out)->npix/2; k++) {
                t = (*out)->data[i][j][k];
                (*out)->data[i][j][k] = (*out)->data[i][(*out)->nlin/2+j][(*out)->npix/2+k];
                (*out)->data[i][(*out)->nlin/2+j][(*out)->npix/2+k] = t;
                t = (*out)->data[i][(*out)->nlin/2+j][k];
                (*out)->data[i][(*out)->nlin/2+j][k] = (*out)->data[i][j][(*out)->npix/2+k];
                (*out)->data[i][j][(*out)->npix/2+k] = t;
            }
        }
    }

    the_end:
    if (a)     free(a);
    if (pwr)   free(pwr);
    if (sine)  free(sine);
    if (tang)  free(tang);
    if (array) free(array);
    if (TIMES) TIMING(T_EXIT);
}
