#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function calculates a least-squares-fit polynomial to a set of      */
/*   spatial control point pairs from two images, a "reference" image with    */
/*   coordinate system (x',y') and a "transformation" image with coordinate   */
/*   system (x,y) to be warped. The polynomial equations are:                 */
/*                                                                            */
/*      x  =    a1                      y  =    b1                            */
/*            + a2 * x'                       + b2 * x'                       */
/*            + a3 * y'                       + b3 * y'                       */
/*            + a4 * x' * y'                  + b4 * x' * y'                  */
/*            + a5 * x' * x'                  + b5 * x' * x'                  */
/*            + a6 * y' * y'                  + b6 * y' * y'                  */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void GEOMCOEF (
double *x,      /*  I   Pointer to an array[npts], containing x-coordinates   */
                /*      of the control points in the "reference" image.       */
double *y,      /*  I   Pointer to an array[npts], containing y-coordinates   */
                /*      of the control points in the "reference" image.       */
double *xp,     /*  I   Pointer to an array[npts], containing x-coordinates   */
                /*      of the control points in the "transformation" image.  */
double *yp,     /*  I   Pointer to an array[npts], containing y-coordinates   */
                /*      of the control points in the "transformation" image.  */
short  npts,    /*  I   Number of control point pairs.                        */
short  nterms,  /*  I   Number of polynomial terms.                           */
double *xc,     /*  O   Pointer to an array[nterms], containing the           */
                /*      x-coefficients of the coordinate transformation.      */
double *yc      /*  O   Pointer to an array[nterms], containing               */
                /*      y-coefficients of the coordinate transformation.      */
/*----------------------------------------------------------------------------*/
) { register uint32_t i, j, k;
    uint64_t total;
    char   msg[SLEN];
    double *a=NULL, *ata=NULL, *atainv=NULL, *xt=NULL, *yt=NULL, err;

    if (TIMES) TIMING(T_GEOMCOEF);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"GEOMCOEF");
    }

    /* check input */
    if (npts < nterms) {
        MESSAGE('E'," Number of points must be equal to or greater than number of polynomial terms.");
        goto the_end;
    } else if (MINTERMS > nterms) {
        sprintf(msg," Number of terms must be equal to or greater than %d.",MINTERMS);
        MESSAGE('E',msg);
        goto the_end;
    } else if (nterms > MAXTERMS) {
        sprintf(msg," Number of terms must be less than or equal to %d.",MAXTERMS);
        MESSAGE('E',msg);
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Number of control points:  %d",npts);
        MESSAGE('I',msg);
        MESSAGE('I'," Control points:");
        MESSAGE('I'," Point #     Line        Pixel         Lineref    Pixelref");
        for (i=0; i<npts; i++) {
#ifndef SADIE_IPT
            sprintf(msg," %4d%15.4e%12.4e%15.4e%12.4e",i+1,y[i],x[i],yp[i],xp[i]);
#else
            sprintf(msg," %4d%15.4e%12.4e%15.4e%12.4e",i+1,y[i]+1.0,x[i]+1.0,yp[i]+1.0,xp[i]+1.0);
#endif
            MESSAGE('I',msg);
        }
        sprintf(msg," Number of polynomial terms:     %d",nterms);
        MESSAGE('I',msg);
        MESSAGE('I'," ...............");
    }

    /* allocate matrices and vectors */
    total = (uint64_t)npts * (uint64_t)nterms * (uint64_t)sizeof(double);
    if (total > (uint64_t)SIZE_MAX  ||  (a=(double *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }
    total = (uint64_t)nterms * (uint64_t)nterms * (uint64_t)sizeof(double);
    if (total > (uint64_t)SIZE_MAX  ||  (ata=(double *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }
    total = (uint64_t)nterms * (uint64_t)nterms * (uint64_t)sizeof(double);
    if (total > (uint64_t)SIZE_MAX  ||  (atainv=(double *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }
    total = (uint64_t)npts * (uint64_t)sizeof(double);
    if (total > (uint64_t)SIZE_MAX  ||  (xt=(double *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }
    total = (uint64_t)npts * (uint64_t)sizeof(double);
    if (total > (uint64_t)SIZE_MAX  ||  (yt=(double *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }

    /* compute matrix a */
    for (i=0; i<npts; i++) {
        switch (nterms) {
        case 6:
            a[i*nterms+5] = yp[i]*yp[i];
        case 5:
            a[i*nterms+4] = xp[i]*xp[i];
        case 4:
            a[i*nterms+3] = xp[i]*yp[i];
        case 3:
            a[i*nterms+2] = yp[i];
            a[i*nterms+1] = xp[i];
            a[i*nterms+0] = 1.0;
        }
    }

    /* compute and invert matrix ata */
    for (i=0; i<nterms; i++) {
        for (j=0; j<nterms; j++) {
            for (ata[i*nterms+j]=0.0,k=0; k<npts; k++) {
                ata[i*nterms+j] += a[k*nterms+i]*a[k*nterms+j];
            }
        }
    }
    INVERT(ata,(uint32_t)nterms,atainv);

    /* compute transformation coefficients */
    for (i=0; i<nterms; i++) {
        for (xt[i]=yt[i]=0.0,j=0; j<npts; j++) {
            xt[i] += a[j*nterms+i]*x[j];
            yt[i] += a[j*nterms+i]*y[j];
        }
    }
    for (i=0; i<nterms; i++) {
        for (xc[i]=yc[i]=0.0,j=0; j<nterms; j++) {
            xc[i] += atainv[i*nterms+j]*xt[j];
            yc[i] += atainv[i*nterms+j]*yt[j];
        }
    }

    /* output transformation coefficients */
    MESSAGE('I',"");
    MESSAGE('I'," Transformation Coefficients:");
    sprintf(msg," Line  =   %12.4e                          Pixel  =   %12.4e",yc[0],xc[0]);
    MESSAGE('I',msg);
    sprintf(msg,"          +%12.4e * pixelref                         +%12.4e * pixelref",yc[1],xc[1]);
    MESSAGE('I',msg);
    sprintf(msg,"          +%12.4e * lineref                          +%12.4e * lineref",yc[2],xc[2]);
    MESSAGE('I',msg);
    if (nterms > 3) {
        sprintf(msg,"          +%12.4e * pixelref * lineref               +%12.4e * pixelref * lineref",yc[3],xc[3]);
        MESSAGE('I',msg);
    }
    if (nterms > 4) {
        sprintf(msg,"          +%12.4e * pixelref * pixelref              +%12.4e * pixelref * pixelref",yc[4],xc[4]);
        MESSAGE('I',msg);
    }
    if (nterms > 5) {
        sprintf(msg,"          +%12.4e * lineref  * lineref               +%12.4e * lineref  * lineref",yc[5],xc[5]);
        MESSAGE('I',msg);
    }

    /* compute and output control point differences */
    MESSAGE('I',"");
    MESSAGE('I'," Goodness-of-fit at Control Points:");
    MESSAGE('I',"               Control Point             Polynomial Fit               Difference");
    MESSAGE('I'," Point #     Line        Pixel          Line        Pixel          dLine      dPixel");
    for (err=0.0,i=0; i<npts; i++) {
        xt[i] = yt[i] = 0.0;
        switch (nterms) {
        case 6:
            xt[i] += xc[5]*yp[i]*yp[i];
            yt[i] += yc[5]*yp[i]*yp[i];
        case 5:
            xt[i] += xc[4]*xp[i]*xp[i];
            yt[i] += yc[4]*xp[i]*xp[i];
        case 4:
            xt[i] += xc[3]*xp[i]*yp[i];
            yt[i] += yc[3]*xp[i]*yp[i];
        case 3:
            xt[i] += xc[0] + xc[1]*xp[i] + xc[2]*yp[i];
            yt[i] += yc[0] + yc[1]*xp[i] + yc[2]*yp[i];
        }
        err += (xt[i]-x[i])*(xt[i]-x[i]) + (yt[i]-y[i])*(yt[i]-y[i]);
#ifndef SADIE_IPT
        sprintf(msg," %4d%15.4e%12.4e%15.4e%12.4e%15.4e%12.4e",i+1,y[i],x[i],yt[i],xt[i],y[i]-yt[i],x[i]-xt[i]);
#else
        sprintf(msg," %4d%15.4e%12.4e%15.4e%12.4e%15.4e%12.4e",i+1,y[i]+1.0,x[i]+1.0,yt[i]+1.0,xt[i]+1.0,y[i]-yt[i],x[i]-xt[i]);
#endif
        MESSAGE('I',msg);
    }
    MESSAGE('I',"");
    sprintf(msg," Root-Mean-Squared Difference at Control Points = %10.4e",sqrt(err/(double)npts));
    MESSAGE('I',msg);

    the_end:
    if (a)      free(a);
    if (ata)    free(ata);
    if (atainv) free(atainv);
    if (xt)     free(xt);
    if (yt)     free(yt);
    if (TIMES)  TIMING(T_EXIT);
}
