#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"
#ifdef SADIE_LIB
#include "app_sadie.h"
#endif /* defined SADIE_LIB */

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function computes a two-dimensional gradient (magnitude and         */
/*   direction) of an image, using two user-supplied convolution kernels.     */
/*   The magnitude is computed as the vector magnitude of the output          */
/*   of the two kernels, and the direction is computed as the angle           */
/*   between the two orthogonal gradient vectors.                             */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Background Information-----------------------------------------------------*/
/*                                                                            */
/*   Robinson, G.S.:                                                          */
/*   "Detection and Coding of Edges Using Directional Masks."                 */
/*   Optical Engineering, Vol. 16, No. 6 (Nov/Dec 1977), pp. 580-585          */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void GRADIENT (
IMAGE *in,      /*  I   Pointer to the input image.                           */
PIXEL *vert,    /*  I   Pointer to a square convolution kernel[size][size] for*/
                /*      the y-derivative. It should return positive values    */
                /*      for gradients increasing to the "top" of the image.   */
PIXEL *horz,    /*  I   Pointer to a square convolution kernel[size][size] for*/
                /*      the x-derivative. It should return positive values    */
                /*      for gradients increasing to the "right" of the image. */
short size,     /*  I   Kernel size in lines and pixels/line.                 */
IMAGE **out1,   /*  O   Address of a pointer to the output image              */
                /*      containing the gradient magnitude.                    */
IMAGE **out2    /*  O   Address of a pointer to the output image              */
                /*      containing the gradient direction.                    */
/*----------------------------------------------------------------------------*/
) { register uint32_t i, j, k, l, m, n=size/2;
    char   msg[SLEN];
    double dv, dh, pinc, psum;
#ifdef SADIE_LIB
    uint32_t progstep, progincr;
    double progtotal;
    uint64_t progcount;
    int progmeter=0;
    int cancel=0;
#endif

#ifdef SADIE_LIB
    progmeter = APP_PROGMETER_CREATE("Gradient");
#endif
    if (TIMES) TIMING(T_GRADIENT);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"GRADIENT");
    }

    /* check input */
    if (!CHECKIMG(in)) {
        MESSAGE('E'," Can't identify image.");
        goto the_end;
    } else if (size < 2  ||  size%2 == 0) {
        MESSAGE('E'," Size of convolution mask must be an odd number greater than two.");
        goto the_end;
    } else if (size > in->nlin  ||  size > in->npix) {
        MESSAGE('E'," Image size must be equal to or greater than convolution mask size.");
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Input image:                    %s",in->text);
        MESSAGE('I',msg);
        sprintf(msg," Kernel size: lines and pixels:  %d",size);
        MESSAGE('I',msg);
        if (size <= SLEN/16) {
            MESSAGE('I'," Horizontal kernel:");
            for (i=0; i<size; i++) {
                sprintf(msg," ");
                for (j=0; j<size; j++) {
                    sprintf(msg+strlen(msg),"%12.4e",(double)horz[i*size+j]);
                }
                MESSAGE('I',msg);
            }
            MESSAGE('I'," Vertical kernel:");
            for (i=0; i<size; i++) {
                sprintf(msg," ");
                for (j=0; j<size; j++) {
                    sprintf(msg+strlen(msg),"%12.4e",(double)vert[i*size+j]);
                }
                MESSAGE('I',msg);
            }
        } else {
            MESSAGE('I'," Kernels too large to list.");
        }
        MESSAGE('I'," ...............");
    }

    /* create images of appropriate size */
    if (!CHECKIMG(*out1)) GETMEM(in->nbnd,in->nlin,in->npix,out1);
    if (!*out1) goto the_end;
    if (!CHECKIMG(*out2)) GETMEM(in->nbnd,in->nlin,in->npix,out2);
    if (!*out2) goto the_end;

    /* initialize progress indicator */
    if (LINES  &&  !PROGRESS(psum=0.0)) goto the_end;
    pinc = 1.0/(double)in->nbnd/(double)(in->nlin-2*n);
#ifdef SADIE_LIB
    if (progmeter) {
        progtotal = (double)in->nbnd * (in->nlin-2*n);
        progcount = 0;
        progstep = progincr = 1 + (uint32_t)floor(progtotal/20);
    }
#endif

    for (i=0; i<in->nbnd; i++) {

        /* compute convolution */
        for (j=n; j<in->nlin-n; j++) {
            for (k=n; k<in->npix-n; k++) {
                dv = dh = 0.0;
                for (l=0; l<size; l++) {
                    for (m=0; m<size; m++) {
                        dv += (double)vert[l*size+m]*(double)in->data[i][j-n+l][k-n+m];
                        dh += (double)horz[l*size+m]*(double)in->data[i][j-n+l][k-n+m];
                    }
                }
                (*out1)->data[i][j][k] = (PIXEL)sqrt((dv*dv+dh*dh));
                if (dh == 0.0) {
                    if (dv > 0.0) {
                        (*out2)->data[i][j][k] = (PIXEL)(PI/2.0);
                    } else if (dv < 0.0) {
                        (*out2)->data[i][j][k] = (PIXEL)(-PI/2.0);
                    } else /* (dv == 0.0 ) */ {
                        (*out2)->data[i][j][k] = (PIXEL)0.0;
                    }
                } else {
                    (*out2)->data[i][j][k] = (PIXEL)atan2(dv,dh);
                }
            }
            if (LINES  &&  !PROGRESS(psum+=pinc)) goto the_end;
#ifdef SADIE_LIB
            if (progmeter && !(--progstep)) {
                progcount += progincr;
                progstep = progincr;
                cancel = APP_PROGMETER_UPDATE(progmeter,(100*progcount/progtotal));
                if (cancel != 0) goto the_end;
            }
#endif
        }

        /* fill top and bottom margins */
        for (j=0; j<n; j++) {
            for (k=n; k<in->npix-n; k++) {
                (*out1)->data[i][j][k]            = (*out1)->data[i][n][k];
                (*out1)->data[i][in->nlin-j-1][k] = (*out1)->data[i][in->nlin-n-1][k];
                (*out2)->data[i][j][k]            = (*out2)->data[i][n][k];
                (*out2)->data[i][in->nlin-j-1][k] = (*out2)->data[i][in->nlin-n-1][k];
            }
        }

        /* fill left and right margins */
        for (j=0; j<in->nlin; j++) {
            for (k=0; k<n; k++) {
                (*out1)->data[i][j][k]            = (*out1)->data[i][j][n];
                (*out1)->data[i][j][in->npix-k-1] = (*out1)->data[i][j][in->npix-n-1];
                (*out2)->data[i][j][k]            = (*out2)->data[i][j][n];
                (*out2)->data[i][j][in->npix-k-1] = (*out2)->data[i][j][in->npix-n-1];
            }
        }
    }

    the_end:
    if (TIMES) TIMING(T_EXIT);
#ifdef SADIE_LIB
    if (progmeter) {
        APP_PROGMETER_DESTROY(progmeter);
    }
    if (cancel != 0) {
        if (*out1) RELMEM(*out1);
        *out1 = NULL;
        if (*out2) RELMEM(*out2);
        *out2 = NULL;
    }
#endif
}
