#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function transforms a hue-saturation-value image                    */
/*   to a red-green-blue image.                                               */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Background Information-----------------------------------------------------*/
/*                                                                            */
/*   Smith, A.R.:                                                             */
/*   "Color Gamut Transform Pairs"                                            */
/*   Computer Graphics, Vol. 12, No. 3 (August 1978), pp. 12-19.              */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void HSVTORGB (
IMAGE *in,      /*  I   Pointer to the input image.                           */
IMAGE **out     /*  O   Address of a pointer to the output image.             */
                /*      The input image may be overwritten (out = &in).       */
/*----------------------------------------------------------------------------*/
) { register uint32_t j, k;
    char   msg[SLEN];
    short  l;
    double h, s, v, f, m, n, i, pinc, psum;

    if (TIMES) TIMING(T_HSVTORGB);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"HSVTORGB");
    }

    /* check input */
    if (!CHECKIMG(in)) {
        MESSAGE('E'," Can't identify image.");
        goto the_end;
    } else if (in->nbnd != 3) {
        MESSAGE('E'," Input image must be a three-band image.");
        goto the_end;
    } else if ((RANGE(in),in->gmin) < (PIXEL)0) {
        MESSAGE('E'," HSV components must be equal to or greater than zero.");
        goto the_end;
    } else if (in->gmax > (PIXEL)SMAX) {
        sprintf(msg," HSV components must be less than or equal to %d.",SMAX);
        MESSAGE('E',msg);
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Input image:   %s",in->text);
        MESSAGE('I',msg);
        MESSAGE('I'," ...............");
    }

    /* create image of appropriate size */
    if (!CHECKIMG(*out)) GETMEM(in->nbnd,in->nlin,in->npix,out);
    if (!*out) goto the_end;

    /* initialize progress indicator */
    if (LINES  &&  !PROGRESS(psum=0.0)) goto the_end;
    pinc = 1.0/(double)in->nlin;

    /* transform HSV to RGB */
    for (j=0; j<in->nlin; j++) {
        for (k=0; k<in->npix; k++) {
            h = 6.0*(double)in->data[0][j][k]/(double)SMAX;
            s =     (double)in->data[1][j][k]/(double)SMAX;
            v =     (double)in->data[2][j][k];
            l = (short)floor(h);
            f = h - (double)l;
            m = v*(1.0-s);
            n = v*(1.0-(s*f));
            i = v*(1.0-(s*(1.0-f)));
            switch (l) {
            case 0:
                (*out)->data[0][j][k] = (PIXEL)v;
                (*out)->data[1][j][k] = (PIXEL)i;
                (*out)->data[2][j][k] = (PIXEL)m;
                break;
            case 1:
                (*out)->data[0][j][k] = (PIXEL)n;
                (*out)->data[1][j][k] = (PIXEL)v;
                (*out)->data[2][j][k] = (PIXEL)m;
                break;
            case 2:
                (*out)->data[0][j][k] = (PIXEL)m;
                (*out)->data[1][j][k] = (PIXEL)v;
                (*out)->data[2][j][k] = (PIXEL)i;
                break;
            case 3:
                (*out)->data[0][j][k] = (PIXEL)m;
                (*out)->data[1][j][k] = (PIXEL)n;
                (*out)->data[2][j][k] = (PIXEL)v;
                break;
            case 4:
                (*out)->data[0][j][k] = (PIXEL)i;
                (*out)->data[1][j][k] = (PIXEL)m;
                (*out)->data[2][j][k] = (PIXEL)v;
                break;
            default:
                (*out)->data[0][j][k] = (PIXEL)v;
                (*out)->data[1][j][k] = (PIXEL)m;
                (*out)->data[2][j][k] = (PIXEL)n;
                break;
            }
        }
        if (LINES  &&  !PROGRESS(psum+=pinc)) goto the_end;
    }

    the_end:
    if (TIMES) TIMING(T_EXIT);
}
