#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function inserts one image into another by replacing overlapping    */
/*   pixels in the "background" image with those from the "foreground" image. */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void INSERT (
IMAGE *in1,     /*  I   Pointer to the background image.                      */
IMAGE *in2,     /*  I   Pointer to the foreground image.                      */
uint32_t jbgn,  /*  I   Line offset in the background image.                  */
uint32_t kbgn,  /*  I   Pixel offset in the background image.                 */
IMAGE **out     /*  O   Address of a pointer to the output image.             */
                /*      The background image may be overwritten (out = &in1). */
/*----------------------------------------------------------------------------*/
) { register uint32_t i, j, k, elin, epix;
    char msg[SLEN];

    if (TIMES) TIMING(T_INSERT);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"INSERT");
    }

    /* check input */
    if (!CHECKIMG(in1)) {
        MESSAGE('E'," Can't identify background image.");
        goto the_end;
    } else if (!CHECKIMG(in2)) {
        MESSAGE('E'," Can't identify foreground image.");
        goto the_end;
    } else if (in1->nbnd != in2->nbnd) {
        MESSAGE('E'," Number of bands must be identical in both images.");
        goto the_end;
    } else if (0 > jbgn  ||  jbgn >= in1->nlin  ||  (uint64_t)jbgn+(uint64_t)in2->nlin > (uint64_t)in1->nlin) {
        MESSAGE('E'," Foreground image must be located completely inside the background image in the line dimension.");
        goto the_end;
    } else if (0 > kbgn  ||  kbgn >= in1->npix  ||  (uint64_t)kbgn+(uint64_t)in2->npix > (uint64_t)in1->npix) {
        MESSAGE('E'," Foreground image must be located completely inside the background image in the pixel dimension.");
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Background input image:   %s",in1->text);
        MESSAGE('I',msg);
        sprintf(msg," Foreground input image:   %s",in2->text);
        MESSAGE('I',msg);
        sprintf(msg," Offset: lines:            %lu",jbgn);
        MESSAGE('I',msg);
        sprintf(msg,"         pixels:           %lu",kbgn);
        MESSAGE('I',msg);
        MESSAGE('I'," ...............");
    }

    /* create image of appropriate size */
    if (!CHECKIMG(*out)) GETMEM(in1->nbnd,in1->nlin,in1->npix,out);
    if (!*out) goto the_end;

    /* copy image */
    for (i=0; i<in1->nbnd; i++) {
        for (j=0; j<in1->nlin; j++) {
            for (k=0; k<in1->npix; k++) {
                (*out)->data[i][j][k] = in1->data[i][j][k];
            }
        }
    }

    /* replace overlapping pixels */
    elin = min(in1->nlin,jbgn+in2->nlin);
    epix = min(in1->npix,kbgn+in2->npix);
    for (i=0; i<in1->nbnd; i++) {
        for (j=jbgn; j<elin; j++) {
            for (k=kbgn; k<epix; k++) {
                (*out)->data[i][j][k] = in2->data[i][j-jbgn][k-kbgn];
            }
        }
    }

    the_end:
    if (TIMES) TIMING(T_EXIT);
}
