#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"

#define         NTIMES          12

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1990 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function creates an image consisting of random noise. The noise     */
/*   either has a uniform distribution with a range of sigma, or it has a     */
/*   Gaussian distribution with a standard deviation of sigma.                */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void RANDOM (
uint32_t nlin,  /*  I   Number of lines in the output image.                  */
uint32_t npix,  /*  I   Number of pixels/line in the output image.            */
short  option,  /*  I   Switch, indicating the noise distribution:            */
                /*      UNIFORM    -   Uniform distribution.                  */
                /*      GAUSSIAN   -   Gaussian distribution.                 */
double sigma,   /*  I   Standard deviation of the noise.                      */
IMAGE  **out    /*  O   Address of a pointer to the output image.             */
/*----------------------------------------------------------------------------*/
) { register uint32_t j, k;
    char msg[SLEN];
    short l;
    double noise, factor, bias;

    if (TIMES) TIMING(T_RANDOM);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"RANDOM");
    }

    /* check input */
    if (nlin <= 0) {
        MESSAGE('E'," Number of lines must be greater than zero.");
        goto the_end;
    } else if (npix <= 0) {
        MESSAGE('E'," Number of pixels/line must be greater than zero.");
        goto the_end;
    } else if (option != UNIFORM  &&  option != GAUSSIAN) {
        MESSAGE('E'," Option must be one of UNIFORM or GAUSSIAN.");
        goto the_end;
    } else if (sigma <= 0.0) {
        MESSAGE('E'," Noise variance must be greater than zero.");
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Image size: lines:    %lu",nlin);
        MESSAGE('I',msg);
        sprintf(msg,"             pixels:   %lu",npix);
        MESSAGE('I',msg);
        if (option == UNIFORM) {
            sprintf(msg," Uniform distribution, range: %12.4e",sigma);
        } else if (option == GAUSSIAN) {
            sprintf(msg," Gaussian distribution, standard deviation: %12.4e",sigma);
        }
        MESSAGE('I',msg);
        MESSAGE('I'," ...............");
    }

    /* create image of appropriate size */
    if (!CHECKIMG(*out)) GETMEM(1,nlin,npix,out);
    if (!*out) goto the_end;
        
    factor = (double)RAND_MAX;
    if (option == UNIFORM) {
        bias = 0.5; 
        for (j=0; j<nlin; j++) {
            for (k=0; k<npix; k++) {

                /* convert uniform distribution to range zero to one */
                noise = (double)rand() / factor;

                /* zero mean and scale to sigma */
                (*out)->data[0][j][k] = (PIXEL)(sigma*(noise - bias));
            }
        }
    } else /* (option == GAUSSIAN) */ {
        bias = (double)NTIMES / 2.0;
        for (j=0; j<nlin; j++) {
            for (k=0; k<npix; k++) {

                /* convert uniform distribution to gaussian over range zero to one */
                for (noise=0.0,l=0; l<NTIMES; l++) {
                    noise += (double)rand();
                }
                noise /= factor;

                /* zero mean and scale to sigma */
                (*out)->data[0][j][k] = (PIXEL)(sigma*(noise - bias));
            }
        }
    }

    the_end:
    if (TIMES) TIMING(T_EXIT);
}

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 2003 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function creates an image consisting of random noise. The noise     */
/*   either has a uniform distribution with a range of sigma, or it has a     */
/*   Gaussian distribution with a standard deviation of sigma.                */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void RANDOM3D (
uint32_t nbnd,  /*  I   Number of bands in the output image.                  */
uint32_t nlin,  /*  I   Number of lines in the output image.                  */
uint32_t npix,  /*  I   Number of pixels/line in the output image.            */
short  option,  /*  I   Switch, indicating the noise distribution:            */
                /*      UNIFORM    -   Uniform distribution.                  */
                /*      GAUSSIAN   -   Gaussian distribution.                 */
double sigma,   /*  I   Standard deviation of the noise.                      */
IMAGE  **out    /*  O   Address of a pointer to the output image.             */
/*----------------------------------------------------------------------------*/
) { register uint32_t i, j, k;
    char msg[SLEN];
    short l;
    double noise, factor, bias;

    if (TIMES) TIMING(T_RANDOM3D);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"RANDOM3D");
    }

    /* check input */
    if (nlin <= 0) {
        MESSAGE('E'," Number of lines must be greater than zero.");
        goto the_end;
    } else if (npix <= 0) {
        MESSAGE('E'," Number of pixels/line must be greater than zero.");
        goto the_end;
    } else if (option != UNIFORM  &&  option != GAUSSIAN) {
        MESSAGE('E'," Option must be one of UNIFORM or GAUSSIAN.");
        goto the_end;
    } else if (sigma <= 0.0) {
        MESSAGE('E'," Noise variance must be greater than zero.");
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Image size: bands:    %lu",nbnd);
        MESSAGE('I',msg);
        sprintf(msg,"             lines:    %lu",nlin);
        MESSAGE('I',msg);
        sprintf(msg,"             pixels:   %lu",npix);
        MESSAGE('I',msg);
        if (option == UNIFORM) {
            sprintf(msg," Uniform distribution, range: %12.4e",sigma);
        } else if (option == GAUSSIAN) {
            sprintf(msg," Gaussian distribution, standard deviation: %12.4e",sigma);
        }
        MESSAGE('I',msg);
        MESSAGE('I'," ...............");
    }

    /* create image of appropriate size */
    if (!CHECKIMG(*out)) GETMEM(nbnd,nlin,npix,out);
    if (!*out) goto the_end;
        
    factor = (double)RAND_MAX;
    if (option == UNIFORM) {
        bias = 0.5; 
        for (i=0; i<nbnd; i++) {
            for (j=0; j<nlin; j++) {
                for (k=0; k<npix; k++) {

                    /* convert uniform distribution to range zero to one */
                    noise = (double)rand() / factor;

                    /* zero mean and scale to sigma */
                    (*out)->data[0][j][k] = (PIXEL)(sigma*(noise - bias));
                }
            }
        }
    } else /* (option == GAUSSIAN) */ {
        bias = (double)NTIMES / 2.0;
        for (i=0; i<nbnd; i++) {
            for (j=0; j<nlin; j++) {
                for (k=0; k<npix; k++) {

                    /* convert uniform distribution to gaussian over range zero to one */
                    for (noise=0.0,l=0; l<NTIMES; l++) {
                        noise += (double)rand();
                    }
                    noise /= factor;

                    /* zero mean and scale to sigma */
                    (*out)->data[i][j][k] = (PIXEL)(sigma*(noise - bias));
                }
            }
        }
    }

    the_end:
    if (TIMES) TIMING(T_EXIT);
}
