#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1990 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function combines two "real" images into one "complex" image.       */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void REAL2COM (
IMAGE *in1,     /*  I   Pointer to the first input image.                     */
IMAGE *in2,     /*  I   Pointer to the second input image.                    */
short option,   /*  I   Switch, indicating the type of real images:           */
                /*      REIMAG   -   real and imaginary images are given.     */
                /*      MAGPHA   -   magnitude and phase images are given.    */
IMAGE **out     /*  O   Address of a pointer to the output image.             */
/*----------------------------------------------------------------------------*/
) { register uint32_t i, j, k, n;
    char msg[SLEN];

    if (TIMES) TIMING(T_REAL2COM);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"REAL2COM");
    }

    /* check input */
    if (!CHECKIMG(in1)) {
        MESSAGE('E'," Can't identify first image.");
        goto the_end;
    } else if (!CHECKIMG(in2)) {
        MESSAGE('E'," Can't identify second image.");
        goto the_end;
    } else if (in1->nbnd != in2->nbnd) {
        MESSAGE('E'," Number of bands must be identical in both images.");
        goto the_end;
    } else if (in1->nlin != in2->nlin) {
        MESSAGE('E'," Number of lines/band must be identical in both images.");
        goto the_end;
    } else if (in1->npix != in2->npix) {
        MESSAGE('E'," Number of pixels/line must be identical in both images.");
        goto the_end;
    } else if ((uint64_t)in1->npix*(uint64_t)2 > (uint64_t)UINT32_MAX) {
        MESSAGE('E'," Image size (number of pixels/line) too large.");
        goto the_end;
    } else if (option != REIMAG  &&  option != MAGPHA) {
        MESSAGE('E'," Option must be either REIMAG or MAGPHA.");
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," First input image:    %s",in1->text);
        MESSAGE('I',msg);
        sprintf(msg," Second input image:   %s",in2->text);
        MESSAGE('I',msg);
        if (option == REIMAG) {
            MESSAGE('I'," Real and imaginary images are given");
        } else if (option == MAGPHA) {
            MESSAGE('I'," Magnitude and phase images are given");
        }
        MESSAGE('I'," ...............");
    }

    /* create image of appropriate size */
    if (!CHECKIMG(*out)) GETMEM(in1->nbnd,in1->nlin,2*in1->npix,out);
    if (!*out) goto the_end;

    if (option == REIMAG) {
        /* combine real and imaginary parts */
        for (i=0; i<in1->nbnd; i++) {
            for (j=0; j<in1->nlin; j++) {
                for (k=0,n=0; k<in1->npix; k+=1,n+=2) {
                    (*out)->data[i][j][n]   = in1->data[i][j][k];
                    (*out)->data[i][j][n+1] = in2->data[i][j][k];
                }
            }
        }
    } else /* (option == MAGPHA) */ {
        /* combine magnitude and phase parts */
        for (i=0; i<in1->nbnd; i++) {
            for (j=0; j<in1->nlin; j++) {
                for (k=0,n=0; k<in1->npix; k+=1,n+=2) {
                    (*out)->data[i][j][n]   = (PIXEL)(double)in1->data[i][j][k]*cos((double)in2->data[i][j][k]);
                    (*out)->data[i][j][n+1] = (PIXEL)(double)in1->data[i][j][k]*sin((double)in2->data[i][j][k]);
                }
            }
        }
    }       

    the_end:
    if (TIMES) TIMING(T_EXIT);
}
