#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function performs a moving window average on an image, and          */
/*   resamples the new image at specified intervals.                          */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void RESAMPL (
IMAGE *in,      /*  I   Pointer to the input image.                           */
uint32_t jsize, /*  I   Number of lines in the window.                        */
uint32_t ksize, /*  I   Number of pixels/line in the window.                  */
uint32_t jinc,  /*  I   Line resampling increment.                            */
uint32_t kinc,  /*  I   Pixels/line resampling increment.                     */
IMAGE **out     /*  O   Address of a pointer to the output image.             */
/*----------------------------------------------------------------------------*/
) { register uint32_t i, j, k, l, m, nlin, npix;
    char   msg[SLEN];
    double subsum, area=(double)jsize*(double)ksize;

    if (TIMES) TIMING(T_RESAMPL);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"RESAMPL");
    }

    /* check input */
    if (!CHECKIMG(in)) {
        MESSAGE('E'," Can't identify image.");
        goto the_end;
    } else if (jsize <= 0) {
        MESSAGE('E'," Number of lines in the averaging window must be greater than zero.");
        goto the_end;
    } else if (in->nlin < jsize) {
        MESSAGE('E'," Image size must be equal to or greater than window size in the line dimension.");
        goto the_end;
    } else if (ksize <= 0) {
        MESSAGE('E'," Number of pixels/line in the averaging window must be greater than zero.");
        goto the_end;
    } else if (in->npix < ksize) {
        MESSAGE('E'," Image size must be equal to or greater than window size in the pixel dimension.");
        goto the_end;
    } else if (jinc <= 0) {
        MESSAGE('E'," Line subsample increment must be greater than zero.");
        goto the_end;
    } else if (kinc <= 0) {
        MESSAGE('E'," Pixel subsample increment must be greater than zero.");
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Input image:                  %s",in->text);
        MESSAGE('I',msg);
        sprintf(msg," Window size: lines:           %lu",jsize);
        MESSAGE('I',msg);
        sprintf(msg,"              pixels:          %lu",ksize);
        MESSAGE('I',msg);
        sprintf(msg," Subsample increment: lines:   %lu",jinc);
        MESSAGE('I',msg);
        sprintf(msg,"                      pixels:  %lu",kinc);
        MESSAGE('I',msg);
        MESSAGE('I'," ...............");
    }

    /* create image of appropriate size */
    nlin = (in->nlin-jsize)/jinc + 1;
    npix = (in->npix-ksize)/kinc + 1;
    if (!CHECKIMG(*out)) GETMEM(in->nbnd,nlin,npix,out);
    if (!*out) goto the_end;

    /* perform moving average and resample */
    for (i=0; i<in->nbnd; i++) {
        for (j=0; j<nlin; j++) {
            for (k=0; k<npix; k++) {
                for (subsum=0.0,l=0; l<jsize; l++) {
                    for (m=0; m<ksize; m++) {
                        subsum += (double)in->data[i][j*jinc+l][k*kinc+m];
                    }
                }
                (*out)->data[i][j][k] = (PIXEL)(subsum/area);
            }
        }
    }

    the_end:
    if (TIMES) TIMING(T_EXIT);
}
