#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function transforms a red-green-blue image                          */
/*   to a hue-saturation-value image.                                         */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Background Information-----------------------------------------------------*/
/*                                                                            */
/*   Smith, A.R.:                                                             */
/*   "Color Gamut Transform Pairs"                                            */
/*   Computer Graphics, Vol. 12, No. 3 (August 1978), pp. 12-19.              */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void RGBTOHSV (
IMAGE *in,      /*  I   Pointer to the input image.                           */
IMAGE **out     /*  O   Address of a pointer to the output image.             */
                /*      The input image may be overwritten (out = &in).       */
/*----------------------------------------------------------------------------*/
) { register uint32_t j, k;
    char   msg[SLEN];
    double r, g, b, h, s, v, pinc, psum;

    if (TIMES) TIMING(T_RGBTOHSV);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"RGBTOHSV");
    }

    /* check input */
    if (!CHECKIMG(in)) {
        MESSAGE('E'," Can't identify image.");
        goto the_end;
    } else if (in->nbnd != 3) {
        MESSAGE('E'," Input image must be a three-band image.");
        goto the_end;
    } else if ((RANGE(in),in->gmin) < (PIXEL)0) {
        MESSAGE('E'," RGB components must be equal to or greater than zero.");
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Input image:   %s",in->text);
        MESSAGE('I',msg);
        MESSAGE('I'," ...............");
    }

    /* create image of appropriate size */
    if (!CHECKIMG(*out)) GETMEM(in->nbnd,in->nlin,in->npix,out);
    if (!*out) goto the_end;

    /* initialize progress indicator */
    if (LINES  &&  !PROGRESS(psum=0.0)) goto the_end;
    pinc = 1.0/(double)in->nlin;

    /* transform RGB to HSV */
    for (j=0; j<in->nlin; j++) {
        for (k=0; k<in->npix; k++) {
            r = (double)in->data[0][j][k];
            g = (double)in->data[1][j][k];
            b = (double)in->data[2][j][k];
            v = (r > g  ?  r : g) > b  ?  (r > g  ?  r : g) : b;
            s = (r < g  ?  r : g) < b  ?  (r < g  ?  r : g) : b;
            (*out)->data[2][j][k] = (PIXEL)v;
            if (v == s) {
                (*out)->data[0][j][k] = j == 0  ?  (PIXEL)SMIN : (*out)->data[0][j][k-1];
                (*out)->data[1][j][k] = (PIXEL)SMIN;
            } else {
                if (r == v) {
                    h = g == s  ?  5.0+(v-b)/(v-s) : 1.0-(v-g)/(v-s);
                } else if (g == v) {
                    h = b == s  ?  1.0+(v-r)/(v-s) : 3.0-(v-b)/(v-s);
                } else /* (b == s) */ {
                    h = r == s  ?  3.0+(v-g)/(v-s) : 5.0-(v-r)/(v-s);
                }
                (*out)->data[0][j][k] = (PIXEL)(h/6.0*(double)SMAX);
                (*out)->data[1][j][k] = (PIXEL)((v-s)/v*(double)SMAX);
            }
        }
        if (LINES  &&  !PROGRESS(psum+=pinc)) goto the_end;
    }

    the_end:
    if (TIMES) TIMING(T_EXIT);
}
