#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function computes the joint two-dimensional histogram               */
/*   (scattergram) between two bands from either the same or from             */
/*   two different images.  The first input image graylevels are              */
/*   along the ordinate and the second input image graylevels are             */
/*   along the abscissa of the scattergram.                                   */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void SCATTER (
IMAGE *in1,     /*  I   Pointer to the first input image.                     */
IMAGE *in2,     /*  I   Pointer to the second input image.                    */
uint32_t bnd1,  /*  I   Band in the first image.                              */
uint32_t bnd2,  /*  I   Band in the second image.                             */
uint32_t inc,   /*  I   Line and pixel subsample increment.                   */
uint32_t nlev,  /*  I   Histogram resolution (and number of lines and         */
                /*      pixels/line in the output image).                     */
IMAGE **out     /*  O   Address of a pointer to the output image.             */
/*----------------------------------------------------------------------------*/
) { register uint32_t i, j, ix, iy;
    char   msg[SLEN];
    double width, pinc, psum;
    PIXEL  gmin, gmax;

    if (TIMES) TIMING(T_SCATTER);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"SCATTER");
    }

    /* check input */
    if (!CHECKIMG(in1)) {
        MESSAGE('E'," Can't identify first image.");
        goto the_end;
    } else if (!CHECKIMG(in2)) {
        MESSAGE('E'," Can't identify second image.");
        goto the_end;
    } else if (0 > bnd1  ||  bnd1 >= in1->nbnd) {
        MESSAGE('E'," Can't identify first band.");
        goto the_end;
    } else if (0 > bnd2  ||  bnd2 >= in2->nbnd) {
        MESSAGE('E'," Can't identify second band.");
        goto the_end;
    } else if (in1->nlin != in2->nlin) {
        MESSAGE('E'," Number of lines must be identical in both images.");
        goto the_end;
    } else if (in1->npix != in2->npix) {
        MESSAGE('E'," Number of pixels/line must be identical in both images.");
        goto the_end;
    } else if (inc <= 0) {
        MESSAGE('E'," Subsample increment must be greater than zero.");
        goto the_end;
    } else if (nlev <= 0) {
        MESSAGE('E'," Number of histogram bins must be greater than zero.");
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
#ifndef SADIE_IPT
        sprintf(msg," First input image:                    %s - band %lu",in1->text,bnd1);
#else
        sprintf(msg," First input image:                    %s - band %lu",in1->text,bnd1+1);
#endif
        MESSAGE('I',msg);
#ifndef SADIE_IPT
        sprintf(msg," Second input image:                   %s - band %lu",in2->text,bnd2);
#else
        sprintf(msg," Second input image:                   %s - band %lu",in2->text,bnd2+1);
#endif
        MESSAGE('I',msg);
        sprintf(msg," Histogram resolution:                 %lu",nlev);
        MESSAGE('I',msg);
        sprintf(msg," Line and pixel subsample increment:   %lu",inc);
        MESSAGE('I',msg);
        MESSAGE('I'," ...............");
    }

    /* create image of appropriate size */
    if (!CHECKIMG(*out)) GETMEM(1,nlev,nlev,out);
    if (!*out) goto the_end;

    /* initialize progress indicator */
    if (LINES  &&  !PROGRESS(psum=0.0)) goto the_end;
    pinc = 1.0/(double)(in1->nlin/inc + (in1->nlin%inc?1:0));

    /* compute min, max, and bandwidth */
    RANGE(in1);
    RANGE(in2);
    gmin  = min(in1->gmin,in2->gmin);
    gmax  = max(in1->gmax,in2->gmax);
    width = (double)(gmax-gmin)/(double)(nlev-1);

    /* initialize scattergram */
    for (i=0; i<nlev; i++) {
        for (j=0; j<nlev; j++) {
            (*out)->data[0][i][j] = (PIXEL)0;
        }
    }

    /* calculate the scattergram */
    for (i=0; i<in1->nlin; i+=inc) {
        for (j=0; j<in1->npix; j+=inc) {
            ix = (uint32_t)((double)(in1->data[bnd1][i][j]-gmin)/width);
            iy = (uint32_t)((double)(in2->data[bnd2][i][j]-gmin)/width);
            (*out)->data[0][ix][iy] += (PIXEL)1;
        }
        if (LINES  &&  !PROGRESS(psum+=pinc)) goto the_end;
    }

    the_end:
    if (TIMES) TIMING(T_EXIT);
}
