#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include "sadie.h"
#ifdef SADIE_LIB
#include "app_sadie.h"
#endif /* defined SADIE_LIB */

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function computes a two-dimensional spatial convolution between     */
/*   a user-supplied spatial kernel and a single-band image.                  */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void SCONVL (
IMAGE *in,      /*  I   Pointer to the input image.                           */
PIXEL *psf,     /*  I   Pointer to the convolution mask matrix[jsize][ksize]. */
uint32_t jsize, /*  I   Number of lines in the convolution mask.              */
uint32_t ksize, /*  I   Number of pixels/line in the convolution mask.        */
IMAGE **out     /*  O   Address of a pointer to the output image.             */
/*----------------------------------------------------------------------------*/
) { register uint32_t j, k, l, m;
    uint32_t jmarg=jsize/2, kmarg=ksize/2;
    uint64_t n;
    char   msg[SLEN];
    short  box;
    double sum, c1, cn, pinc, psum;
#ifdef SADIE_LIB
    uint32_t progstep, progincr;
    double progtotal;
    uint64_t progcount;
    int progmeter=0;
    int cancel=0;
#endif

#ifdef SADIE_LIB
    progmeter = APP_PROGMETER_CREATE("Convolution");
#endif
    if (TIMES) TIMING(T_SCONVL);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"SCONVL");
    }

    /* check input */
    if (!CHECKIMG(in)) {
        MESSAGE('E'," Can't identify image.");
        goto the_end;
    } else if (in->nbnd > 1) {
        MESSAGE('E'," Image must be single-band.");
        goto the_end;
    } else if (jsize < 0  ||  jsize%2 == 0) {
        MESSAGE('E'," Number of lines in the convolution mask must be an odd number greater than zero.");
        goto the_end;
    } else if (jsize > in->nlin) {
        MESSAGE('E'," Image size must be equal to or greater than convolution mask size in the line dimension.");
        goto the_end;
    } else if (ksize < 0  ||  ksize%2 == 0) {
        MESSAGE('E'," Number of pixels/line in the convolution mask must be an odd number greater than zero.");
        goto the_end;
    } else if (ksize > in->npix) {
        MESSAGE('E'," Image size must be equal to or greater than convolution mask size in the pixel dimension.");
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Input image:           %s",in->text);
        MESSAGE('I',msg);
        sprintf(msg," Kernel size: lines:    %lu",jsize);
        MESSAGE('I',msg);
        sprintf(msg,"              pixels:   %lu",ksize);
        MESSAGE('I',msg);
        if (jsize <= SLEN/16  &&  ksize <= SLEN/16) {
            MESSAGE('I'," Spatial kernel:");
            for (n=j=0; j<jsize; j++) {
                sprintf(msg," ");
                for (k=0; k<ksize; k++) {
                    sprintf(msg+strlen(msg),"%12.4e",(double)psf[n++]);
                }
                MESSAGE('I',msg);
            }
        } else {
            MESSAGE('I'," Kernel too large to list.");
        }
        MESSAGE('I'," ...............");
    }

    /* create image of appropriate size */
    if (!CHECKIMG(*out)) GETMEM(1,in->nlin,in->npix,out);
    if (!*out) goto the_end;

    /* initialize progress indicator */
    if (LINES  &&  !PROGRESS(psum=0.0)) goto the_end;
    pinc = 1.0/(double)(in->nlin-2*jmarg);
#ifdef SADIE_LIB
    if (progmeter) {
        progtotal = in->nlin-2*jmarg;
        progcount = 0;
        progstep = progincr = 1 + (uint32_t)floor(progtotal/20);
    }
#endif

    /* see if it's a box-filter PSF */
    for (box=TRUE,n=1; box  &&  n < (uint64_t)jsize*(uint64_t)ksize; n++) {
        box = psf[n] == psf[0];
    }
    
    if (box) {
        for (j=jmarg; j<in->nlin-jmarg; j++) {

            /* initialize box */
            for (c1=sum=0.0,l=j-jmarg; l<=j+jmarg; l++) {
                c1 += (double)in->data[0][l][0];
                for (k=1; k<ksize; k++) {
                    sum += (double)in->data[0][l][k];
                }
            }
            (*out)->data[0][j][kmarg] = (PIXEL)((sum+=c1)*(double)psf[0]);

            /* box-filter algorithm */
            for (k=kmarg+1; k<in->npix-kmarg; k++) {
                for (sum-=c1,c1=cn=0.0,l=j-jmarg; l<=j+jmarg; l++) {
                    c1 += (double)in->data[0][l][k-kmarg];
                    cn += (double)in->data[0][l][k+kmarg];
                }
                (*out)->data[0][j][k] = (PIXEL)((sum+=cn)*(double)psf[0]);
            }
            if (LINES  &&  !PROGRESS(psum+=pinc)) goto the_end;
#ifdef SADIE_LIB
            if (progmeter && !(--progstep)) {
                progcount += progincr;
                progstep = progincr;
                cancel = APP_PROGMETER_UPDATE(progmeter,(100*progcount/progtotal));
                if (cancel != 0) goto the_end;
            }
#endif
        }

    } else /* compute normal convolution */ {
        for (j=jmarg; j<in->nlin-jmarg; j++) {
            for (k=kmarg; k<in->npix-kmarg; k++) {
                for (sum=0.0,n=(uint64_t)jsize*(uint64_t)ksize,l=0; l<jsize; l++) {
                    for (m=0; m<ksize; m++) {
                        sum += (double)in->data[0][j-jmarg+l][k-kmarg+m]*(double)psf[--n];
                    }
                }
                (*out)->data[0][j][k] = (PIXEL)sum;
            }
            if (LINES  &&  !PROGRESS(psum+=pinc)) goto the_end;
#ifdef SADIE_LIB
            if (progmeter && !(--progstep)) {
                progcount += progincr;
                progstep = progincr;
                cancel = APP_PROGMETER_UPDATE(progmeter,(100*progcount/progtotal));
                if (cancel != 0) goto the_end;
            }
#endif
        }
    }

    /* fill top and bottom margins */
    for (j=0; j<jmarg; j++) {
        for (k=0; k<in->npix-kmarg; k++) {
            (*out)->data[0][j][k]            = (*out)->data[0][jmarg][k];
            (*out)->data[0][in->nlin-j-1][k] = (*out)->data[0][in->nlin-jmarg-1][k];
        }
    }

    /* fill left and right margins */
    for (j=0; j<in->nlin; j++) {
        for (k=0; k<kmarg; k++) {
            (*out)->data[0][j][k]            = (*out)->data[0][j][kmarg];
            (*out)->data[0][j][in->npix-k-1] = (*out)->data[0][j][in->npix-kmarg-1];
        }
    }

    the_end:
    if (TIMES) TIMING(T_EXIT);
#ifdef SADIE_LIB
    if (progmeter) {
        APP_PROGMETER_DESTROY(progmeter);
    }
    if (cancel != 0) {
        if (*out) RELMEM(*out);
    }
#endif
}
