#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1990 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function creates a class signature map from the original image      */
/*   and a class label map produced by a classification or segmentation       */
/*   function.  The signature map consists of the average graylevel for       */
/*   each labeled segment in the image.                                       */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void SIGMAP (
IMAGE  *in1,    /*  I   Pointer to the classification label map.              */
IMAGE  *in2,    /*  I   Pointer to the original image.                        */
IMAGE  **out    /*  O   Address of a pointer to the output signature map.     */
/*----------------------------------------------------------------------------*/
) { register uint32_t i, j, k;
    uint64_t total, n, nlev, index, *count=NULL;
    char  msg[SLEN];
    PIXEL *table=NULL;

    if (TIMES) TIMING(T_SIGMAP);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"SIGMAP");
    }

    if (!CHECKIMG(in1)) {
        MESSAGE('E'," Can't identify image.");
        goto the_end;
    } else if (!CHECKIMG(in2)) {
        MESSAGE('E'," Can't identify image.");
        goto the_end;
    } else if (in1->nbnd != 1) {
        MESSAGE('E'," Classification label map must be single-band.");
        goto the_end;
    } else if (in1->nlin != in2->nlin) {
        MESSAGE('E'," Number of lines must be identical in all images.");
        goto the_end;
    } else if (in1->npix != in2->npix) {
        MESSAGE('E'," Number of pixels/line must be identical in all images.");
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Label map:        %s",in1->text);
        MESSAGE('I',msg);
        sprintf(msg," Original image:   %s",in2->text);
        MESSAGE('I',msg);
        MESSAGE('I'," ...............");
    }

    /* create image of appropriate size */
    if (!CHECKIMG(*out)) GETMEM(in2->nbnd,in2->nlin,in2->npix,out);
    if (!*out) goto the_end;

    /* allocate and initialize accumulator arrays */
    RANGE(in1);
    nlev  = (uint64_t)(in1->gmax - in1->gmin + 1);
    total = (uint64_t)nlev * (uint64_t)sizeof(uint64_t);
    if (total > (uint64_t)SIZE_MAX  ||  (count=(uint64_t *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }
    total = (uint64_t)nlev * (uint64_t)in2->nbnd * (uint64_t)sizeof(PIXEL);
    if (total > (uint64_t)SIZE_MAX  ||  (table=(PIXEL *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }
    for (n=0; n<nlev; n++) {
        count[n] = 0;
    }
    for (n=0; n<nlev*(uint64_t)in2->nbnd; n++) {
        table[n] = (PIXEL)0;
    }

    /* first pass: accumulate segment graylevels */
    for (j=0; j<in2->nlin; j++) {
        for (k=0; k<in2->npix; k++) {
            index = (uint64_t)(in1->data[0][j][k] - in1->gmin);
            count[index] += 1;
            for (i=0; i<in2->nbnd; i++) {
                table[(uint64_t)i*nlev+index] += in2->data[i][j][k];
            }
        }
    }

    /* calculate average segment graylevels */
    for (n=0; n<nlev; n++) {
        if (count[n] > 0) {
            for (i=0; i<in2->nbnd; i++) {
                table[(uint64_t)i*nlev+n] /= (PIXEL)count[n];
            }
        }
    }

    /* second pass: replace segment labels with average graylevels */
    for (j=0; j<in2->nlin; j++) {
        for (k=0; k<in2->npix; k++) {
            index = (uint64_t)(in1->data[0][j][k] - in1->gmin);
            for (i=0; i<in2->nbnd; i++) {
                (*out)->data[i][j][k] = table[(uint64_t)i*nlev+index];
            }
        }
    }

    the_end:
    if (count) free(count);
    if (table) free(table);
    if (TIMES) TIMING(T_EXIT);
}
