#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function performs a spatially-variable graylevel transformation     */
/*   on a single-band image.                                                  */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Background Information-----------------------------------------------------*/
/*                                                                            */
/*   Fahnestock, J.D. and Schowengerdt, R.A.:                                 */
/*   "Spatially variant contrast enhancement using local range modification." */
/*   Optical Engineering, Vol. 22, No. 3, May/June 1983                       */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void VSTRETCH (
IMAGE *in,      /*  I   Pointer to the input image.                           */
uint32_t size,  /*  I   Number of lines and pixels in image blocks.           */
IMAGE **out     /*  O   Address of a pointer to the output image.             */
/*----------------------------------------------------------------------------*/
) { register uint32_t j, k, u, v;
    uint64_t total, x, y, idx1, idx2, blkx, blky;
    char   msg[SLEN];
    double rat1, rat2, rat3, rat4, pinc, psum;
    PIXEL  *hi=NULL, *lo=NULL, *mp=NULL, *np=NULL, m, n;

    if (TIMES) TIMING(T_VSTRETCH);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"VSTRETCH");
        MESSAGE('I',"");
        sprintf(msg," Input image:                    %s",in->text);
        MESSAGE('I',msg);
        sprintf(msg," Block size: lines and pixels:   %lu",size);
        MESSAGE('I',msg);
        MESSAGE('I'," ...............");
    }

    /* check input */
    if (!CHECKIMG(in)) {
        MESSAGE('E'," Can't identify image.");
        goto the_end;
    } else if (in->nbnd > 1) {
        MESSAGE('E'," Image must be single-band.");
        goto the_end;
    } else if (size < 1) {
        MESSAGE('E'," Size of subblock must be equal to or greater than one.");
        goto the_end;
    } else if (size > in->nlin  ||  size > in->npix) {
        MESSAGE('E'," Image size must be equal to or greater than subblock size.");
        goto the_end;
    }

    /* create image of appropriate size */
    if (!CHECKIMG(*out)) GETMEM(1,in->nlin,in->npix,out);
    if (!*out) goto the_end;
    blkx = (in->npix-1)/size + 1;
    blky = (in->nlin-1)/size + 1;

    /* allocate space for arrays */
    total = (uint64_t)blkx * (uint64_t)blky * (uint64_t)sizeof(PIXEL);;
    if (total > (uint64_t)SIZE_MAX  ||  (hi=(PIXEL *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }
    total = (uint64_t)blkx * (uint64_t)blky * (uint64_t)sizeof(PIXEL);;
    if (total > (uint64_t)SIZE_MAX  ||  (lo=(PIXEL *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }
    total = (uint64_t)(blkx+1) * (uint64_t)(blky+1) * (uint64_t)sizeof(PIXEL);
    if (total > (uint64_t)SIZE_MAX  ||  (mp=(PIXEL *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }
    total = (uint64_t)(blkx+1) * (uint64_t)(blky+1) * (uint64_t)sizeof(PIXEL);
    if (total > (uint64_t)SIZE_MAX  ||  (np=(PIXEL *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }

    /* initialize progress indicator */
    if (LINES  &&  !PROGRESS(psum=0.0)) goto the_end;
    pinc = 0.5/(double)blkx/(double)blky;

    /* first pass: obtain transformation parameters from subblocks */
    /* search for maximum and minimum */
    for (y=0; y<blky; y++) {
        for (x=0; x<blkx; x++) {
            for (m=n=in->data[0][y*size][x*size],j=y*size; j<(y+1)*size && j<in->nlin; j++) {
                for (k=x*size; k<(x+1)*size && k<in->npix; k++) {
                    if (in->data[0][j][k] > m) m = in->data[0][j][k];
                    if (in->data[0][j][k] < n) n = in->data[0][j][k];
                }
            }
            hi[y*blkx+x] = m;
            lo[y*blkx+x] = n;
            if (LINES  &&  !PROGRESS(psum+=pinc)) goto the_end;
        }
    }
    
    /* obtain values for parameter arrays */
    /* four corners */
    mp[0] = hi[0];
    np[0] = lo[0];
    mp[blkx] = hi[blkx-1];
    np[blkx] = lo[blkx-1];
    mp[blky*(blkx+1)] = hi[(blky-1)*blkx];
    np[blky*(blkx+1)] = lo[(blky-1)*blkx];
    mp[(blkx+1)*(blky+1)-1] = hi[blkx*blky-1];
    np[(blkx+1)*(blky+1)-1] = lo[blkx*blky-1];
    
    /* four sides */
    for (y=1; y<blky; y++) {
        idx1 = (y-1)*blkx;
        idx2 = y*blkx;
        mp[y*(blkx+1)] = max(hi[idx1],hi[idx2]);
        np[y*(blkx+1)] = min(lo[idx1],lo[idx2]);
        idx1 = (y+1)*blkx - 1;
        mp[(y+1)*(blkx+1)-1] = max(hi[idx2-1],hi[idx1]);
        np[(y+1)*(blkx+1)-1] = min(lo[idx2-1],lo[idx1]);
    }
    for (x=1; x<blkx; x++) {
        mp[x] = max(hi[x-1],hi[x]);
        np[x] = min(lo[x-1],lo[x]);
        idx1 = blky*(blkx+1)+x;
        idx2 = (blky-1)*blkx+x;
        mp[idx1] = max(hi[idx2-1],hi[idx2]);
        np[idx1] = min(lo[idx2-1],lo[idx2]);
    }
    
    /* inside */
    for (y=1; y<blky; y++) {
        for (x=1; x<blkx; x++) {
            idx1 = (y-1)*blkx + x;
            idx2 = y*blkx + x;
            mp[y*(blkx+1)+x] = max(max(hi[idx1],hi[idx1-1]),max(hi[idx2],hi[idx2-1]));
            np[y*(blkx+1)+x] = min(min(lo[idx1],lo[idx1-1]),min(lo[idx2],lo[idx2-1]));
        }
    }

    /* second pass: perform transformation */
    for (x=0; x<blkx; x++) {
        for (y=0; y<blky; y++) {
            for (u=0,j=y*size; j<(y+1)*size && j<in->nlin; u++,j++) {
                for (v=0,k=x*size; k<(x+1)*size && k<in->npix; v++,k++) {
                    idx1 = (y+1)*(blkx+1)+x;
                    idx2 = y*(blkx+1)+x;
                    rat1 = (double)u/(double)size;
                    rat2 = (double)(size-u)/(double)size;
                    rat3 = (double)(size-v)/(double)size;
                    rat4 = (double)v/(double)size;
                    m = (rat1*(double)mp[idx1]+rat2*(double)mp[idx2])*rat3 + (rat1*(double)mp[idx1+1]+rat2*(double)mp[idx2+1])*rat4;
                    n = (rat1*(double)np[idx1]+rat2*(double)np[idx2])*rat3 + (rat1*(double)np[idx1+1]+rat2*(double)np[idx2+1])*rat4;
                    (*out)->data[0][j][k] = m == n ? in->data[0][j][k] : (PIXEL)((double)(in->gmax-in->gmin)/(double)(m-n)*((double)in->data[0][j][k]-n)) + in->gmin;
                }
            }
            if (LINES  &&  !PROGRESS(psum+=pinc)) goto the_end;
        }
    }

    the_end:
    if (hi)    free(hi);
    if (lo)    free(lo);
    if (mp)    free(mp);
    if (np)    free(np);
    if (TIMES) TIMING(T_EXIT);
}
